from wsgiref.simple_server import make_server
import logging
import click

from falcon_helpers.contrib.wsgi import CustomLoggingWSGIRequestHandler

log = logging.getLogger(__name__)


@click.group()
def server():
    """Basic development server

    Add this to your base command application command like this

        @click.group()
        @click.pass_context
        @click.option('-c', '--config', type=click.Path(exists=True), default='config.ini')
        def app(ctx, config):
            ctx.meta['app'] = app.create_app(config)


        app.add_command(server)
    """
    pass


@server.command()
@click.option('-h', '--host', type=str, default='127.0.0.1')
@click.option('-p', '--port', type=int, default=9090)
@click.pass_context
def start(ctx, host, port):
    try:
        app = ctx.meta['app']
    except KeyError:
        raise click.ClickException(
            'Unable to find an application in click context, looking for meta[\'app\']'
        )

    log.info(f'Starting the dev server on {host}:{port} this should not be used in production...')

    httpd = make_server(host, port, app, handler_class=CustomLoggingWSGIRequestHandler)
    httpd.serve_forever()



import logging

import falcon
import requests

from ..middlewares.jinja2 import Jinja2Response


log = logging.getLogger(__name__)


def _default_failed(req, resp, error, message, **kwargs):
    raise falcon.HTTPTemporaryRedirect('/')


class LoginFormResource:
    auth_required = False

    def __init__(self, client_url, client_id, callback_uri, template_path='auth/login.html'):
        self.client_url = client_url
        self.client_id = client_id
        self.callback_uri = callback_uri
        self.template_path = template_path

    # Load form
    def on_get(self, req, resp):
        resp.context['template'] = Jinja2Response(
            self.template_path,
            auth0_config={
                'client_url': self.client_url,
                'client_id': self.client_id,
                'callback_uri': self.callback_uri,
            })


class LoginCallbackResource:
    auth_required = False

    def __init__(self, client_url, client_id, client_secret, callback_uri,
                 cookie_domain, secure_cookie, cookie_name='X-AuthToken',
                 cookie_max_age=(24 * 60 * 60), cookie_path='/',
                 redirect_uri='/dashboard', after_login=None, when_fails=_default_failed,
                 ):
        self.client_url = client_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.callback_uri = callback_uri
        self.redirect_uri = redirect_uri
        self.failed_action = when_fails

        self.cookie_domain = cookie_domain

        if not isinstance(secure_cookie, bool):
            raise RuntimeError(
                f'secure_cookie must be a literal True or False, got {secure_cookie}.'
            )

        self.secure_cookie = secure_cookie
        self.cookie_name = cookie_name
        self.cookie_max_age = cookie_max_age
        self.cookie_path = cookie_path
        self.after_login = after_login or (lambda data, req, resp: resp)

    def on_get(self, req, resp):
        auth_code = req.params.get('code', None)

        if auth_code is None:
            error = req.params.get('error', None)

            if error is None:
                # Something wierd is happening, likely someone is hitting our callback without it
                # coming from Auth0, in that case, log and run away.
                msg = (
                    f'The callback url was reached without an `auth_code` and without and `error` '
                    f'in the params. That should not be possible if it is coming from Auth0.'
                )
                log.warning(msg)
                self.failed_action(req, resp, error='no_auth_token_or_error', message=msg)
            else:
                msg = req.params.get('error_description', '')
                self.failed_action(req, resp, error=error, message=msg)

            self.failed_action(req, resp, error='no_auth_token',
                               message='No Authentication token was found for this request.')

            # failed_action might not raise and might not eject us from this request, if that
            # happens, we can't continue so through a bad request.
            log.error(
                f'Someone accessed the authentication callback url and did not have an error or an '
                f'auth_code. We called failed_action, but you did not raise an exception. We can '
                f'not continue with this request. Failed actions should handle the error.'
            )
            raise falcon.HTTPInternalServerError(
                title='Authentication Error',
                description='An internal error occurred during authentication.'
            )

        data = self.get_user_data(auth_code)

        self.set_cookie_contents(resp, data)
        resp.set_header('Location', self.redirect_uri)
        resp.status = falcon.HTTP_302

        self.after_login(data, req=req, resp=resp)

    def get_user_data(self, auth_code):
        token = requests.post(
            f'https://{self.client_url}/oauth/token',
            data={
                'grant_type': 'authorization_code',
                'client_id': self.client_id,
                'client_secret': self.client_secret,
                'code': auth_code,
                'redirect_uri': self.callback_uri,
            })

        if not token.ok:
            raise falcon.HTTPTemporaryRedirect('/auth/login')

        return token.json()

    def set_cookie_contents(self, resp, data):
        resp.set_cookie(self.cookie_name,
                        data['id_token'],
                        max_age=self.cookie_max_age,
                        secure=self.secure_cookie,
                        path=self.cookie_path,
                        domain=self.cookie_domain)


class LogoutResource:
    auth_required = False

    def __init__(self, cookie_domain, secure_cookie, cookie_name='X-AuthToken',
                 cookie_path='/'):
        self.cookie_domain = cookie_domain
        self.secure_cookie = secure_cookie

        self.cookie_name = cookie_name
        self.cookie_path = cookie_path

    def on_get(self, req, resp):
        resp.set_cookie(self.cookie_name, '', secure=self.secure_cookie,
                        path=self.cookie_path, domain=self.cookie_domain)

        resp.status = falcon.HTTP_302
        resp.set_header('Location', '/auth/login')



import mimetypes
import random
import string


_BOUNDARY_CHARS = string.digits + string.ascii_letters


def encode_multipart(fields, files, boundary=None):
    """Create a multi-part http request

    Original was found here with an MIT license:
        http://code.activestate.com/recipes/578668-encode-multipart-form-data-for-uploading-files-via/

    Encode dict of form fields and dict of files as multipart/form-data. Return tuple of
    (body_string, headers_dict).

    Each value in files is a dict with required keys 'filename' and 'content', and optional
    'mimetype' (if not specified, tries to guess mime type or uses 'application/octet-stream').
    """

    def escape_quote(s):
        return s.replace('"', '\\"')

    if boundary is None:
        boundary = ''.join(random.choice(_BOUNDARY_CHARS) for i in range(30))

    lines = []

    for name, value in fields.items():
        lines.extend((
            f'--{boundary}',
            f'Content-Disposition: form-data; name="{escape_quote(name)}"',
            '',
            value,
        ))

    for name, value in files.items():
        filename = value['filename']
        mimetype = (value.get('mimetype') or
                    mimetypes.guess_type(filename)[0] or
                    'application/octet-stream')
        name, filename = escape_quote(name), escape_quote(filename)

        lines.extend((
            f'--{boundary}',
            f'Content-Disposition: form-data; name="{name}"; filename="{filename}"',
            f'Content-Type: {mimetype}',
            '',
            value['content'],
        ))

    lines.extend((
        f'--{boundary}--',
        '',
    ))
    body = '\r\n'.join(lines)

    headers = {
        'Content-Type': f'multipart/form-data; boundary={boundary}',
        'Content-Length': str(len(body)),
    }

    return (body, headers)



import pytest

import falcon_helpers.app as fhapp
from falcon_helpers.middlewares.multi import MultiMiddleware


def test_enables_dynamic_middleware():

    app = fhapp.API()
    assert isinstance(app._middleware[0][0].__self__, MultiMiddleware)
    assert isinstance(app._middleware[1][0].__self__, MultiMiddleware)
    assert isinstance(app._middleware[2][0].__self__, MultiMiddleware)

    with pytest.raises(RuntimeError):
        app = fhapp.API(independent_middleware=False)

    app = fhapp.API(enable_dynamic_mw=False)
    assert app._middleware == ((), (), (),)



import unittest.mock as mock
import falcon.testing
import pytest

from falcon_helpers import API
from falcon_helpers.resources.auth0 import LoginCallbackResource


@pytest.fixture()
def app():
    return API()


@pytest.fixture()
def client(app):
    return falcon.testing.TestClient(app)


class TestLoginCallbackResource:

    def test_callback_resource_default_failed(self, client):
        resc = LoginCallbackResource('url', 'id', 'secret', 'uri', 'domain', False)
        client.app.add_route('/auth/callback', resc)
        resp = client.simulate_get('/auth/callback')

        assert resp.status_code == 307
        assert resp.headers['location'] == '/'

    def test_callback_resource_custom_failed(self, client):
        def when_fails(req, resp, error, message, **kwargs):
            resp.media = {'error': 'custom'}
            raise falcon.HTTPStatus('200 OK')

        resc = LoginCallbackResource('url', 'id', 'secret', 'uri', 'domain', False,
                                     when_fails=when_fails)
        client.app.add_route('/auth/callback', resc)
        resp = client.simulate_get('/auth/callback')

        assert resp.status_code == 200
        assert resp.json == {'error': 'custom'}

    def test_callback_resource_custom_failed_doesnt_raise(self, client):
        def when_fails(req, resp, error, message, **kwargs):
            resp.media = {'error': 'custom'}

        resc = LoginCallbackResource('url', 'id', 'secret', 'uri', 'domain', False,
                                     when_fails=when_fails)
        client.app.add_route('/auth/callback', resc)
        resp = client.simulate_get('/auth/callback')

        assert resp.status_code == 500

    @mock.patch('falcon_helpers.resources.auth0.requests.post')
    def test_callback_resource_happy_path(self, m_request, client):
        resc = LoginCallbackResource('url', 'id', 'secret', 'uri', 'domain', False)
        client.app.add_route('/auth/callback', resc)
        resp = client.simulate_get('/auth/callback', params={'code': 'thing'})

        assert resp.status_code == 302

    def test_callback_resource_with_error_code(self, client):
        def when_fails(req, resp, error, message, **kwargs):
            resp.media = {'error': error, 'message': message}
            raise falcon.HTTPStatus('200 OK')

        resc = LoginCallbackResource('url', 'id', 'secret', 'uri', 'domain', False,
                                     when_fails=when_fails)
        client.app.add_route('/auth/callback', resc)
        resp = client.simulate_get('/auth/callback', params={
            'error': 'error',
            'error_description': 'description'
        })

        assert resp.status_code == 200
        assert resp.json == {'error': 'error', 'message': 'description'}



import typing
import sqlalchemy as sa

import falcon_helpers.sqla.db as db
from falcon_helpers.middlewares.sqla import SQLAlchemySessionMiddleware


class SQLAlchemyPlugin:
    """Setup SQLAlchemy for use within an application

    If the application has a config property, that property is a dictionary, and that dictionary has
    a key called `sqla` which has the standard SQLA Engine properties, all you have to do is call
    the `register` function and this plugin will do everything for you.

    Quick start:

        import falcon_helpers.plugins as plugins

        sqla = plugins.SQLAlchemyPlugin()

        class MyApp(falcon_helpers.API):
            pass


        app = MyApp()

        # consider using the configuration class as a part of Falcon-Helpers
        app.config = {
            'sqla': {
                'drivername': '',
                'username': 'myuser',
                'password': 'somepassword',
                'host': 'localhost',
                'port': '5432',
                database': 'myapp'
            }
        }

        sqla.register(app)


    """

    __slots__ = (
        '_url',
        'engine',
        'session',
        'mw',
    )

    def __init__(self, url=None, session=db.session):
        if url:
            self.url = url
        else:
            self._url = None

        self.session = session
        self.engine = None
        self.mw = None

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, url):
        if isinstance(url, typing.Dict):
            self._url = sa.engine.url.URL(**url)
        elif isinstance(url, str):
            self._url = sa.engine.url.make_url(url)
        elif isinstance(url, sa.engine.url.URL):
            self._url = url
        else:
            raise RuntimeError(
                'Passed SQLAlchemy URL does not seem to be a valid value. Expecting dict, string '
                'or url, got {type(url)}'
            )

    def register(self, app):
        if not self.url:
            self.url = app.config.get('sqla', None)

        self.engine = sa.create_engine(self.url)

        if isinstance(self.session, sa.orm.session.Session):
            self.session.bind = self.engine
        else:
            self.session.configure(bind=self.engine)

        self.mw = SQLAlchemySessionMiddleware()
        app.add_middleware(self.mw)

        app.plugins['sqla'] = self



import sqlalchemy as sa

from falcon_helpers.sqla.orm import BaseFunctions
from falcon_helpers.tests.fixtures import Base


def test_column_names():

    class Entity(Base, BaseFunctions):
        __tablename__ = 'entity'
        c1 = sa.Column(sa.Integer, primary_key=True)

    assert Entity.orm_column_names() == {
        'c1',
    }



import jwt
import falcon_helpers.contrib.auth as auth


def test_generate_auth_token():
    # we don't need to hit the db for this test so don't commit it
    u1 = auth.User.testing_create(_commit=False)

    token = u1.generate_auth_token('aud', 'secret')
    result = jwt.decode(token, 'secret', algorithms=['HS512'], audience='aud')
    assert result['sub'] == u1.ident

    token = u1.generate_auth_token('aud', 'secret', additional_data={
        'other': 'true'
    })
    result = jwt.decode(token, 'secret', algorithms=['HS512'], audience='aud')
    assert result['sub'] == u1.ident
    assert result['other'] == 'true'



import pathlib
import falcon_helpers.middlewares as mws
import falcon_helpers.resources.auth0 as rescs


BASEDIR = pathlib.Path(__file__).parent


class Auth0Plugin:

    __slots__ = (
        'callback_url',
        'client_id',
        'client_pubkey',
        'client_secret',
        'client_url',
        'cookie_domain',
        'cookie_name',
        'use_secure_cookie',
        'success_url',
        'login_route',
        'logout_route',
        'callback_route',
        'include_jinja2_middleware',
        'include_auto_parse_jwt',
        'template_dpath',
    )

    def __init__(self, client_url, client_id, client_pubkey, client_secret, callback_url,
                 **options):
        self.client_url = client_url
        self.client_id = client_id
        self.client_pubkey = client_pubkey
        self.client_secret = client_secret
        self.callback_url = callback_url

        self.login_route = options.get('login_route', '/auth/login')
        self.logout_route = options.get('logout_route', '/auth/logout')
        self.callback_route = options.get('callback_route', '/auth/callback')
        self.cookie_domain = options.get('cookie_domain', '')
        self.cookie_name = options.get('cookie_name', 'X-AuthToken')

        self.success_url = options.get('success_url', '/')
        self.use_secure_cookie = options.get('use_secure_cookie', True)
        self.include_jinja2_middleware = options.get('include_jinja2_middleware', True)
        self.include_auto_parse_jwt = options.get('include_auto_parse_jwt', True)
        self.template_dpath = options.get(
            'template_dpath', BASEDIR.joinpath('templates').as_posix()
        )

    def register(self, app):
        self.setup_routes(app)

        if self.include_jinja2_middleware:
            app.add_middleware(mws.Jinja2Middleware(self.template_dpath))

        if self.include_auto_parse_jwt:
            app.add_middleware(
                mws.ParseJWTMiddleware(
                    audience=self.client_id,
                    pubkey=self.client_pubkey,
                    secret=self.client_secret,
                    cookie_name=self.cookie_name,
                )
            )

        app.plugins['auth0'] = self

    def setup_routes(self, app):
        app.add_route(
            self.login_route,
            rescs.LoginFormResource(
                client_url=self.client_url,
                client_id=self.client_id,
                callback_uri=self.callback_url,
                template_path='login.html',
            )
        )
        app.add_route(
            self.logout_route,
            rescs.LogoutResource(
                cookie_domain=self.cookie_domain,
                secure_cookie=self.use_secure_cookie,
            )
        )
        app.add_route(
            '/auth/callback',
            rescs.LoginCallbackResource(
                client_url=self.client_url,
                client_id=self.client_id,
                client_secret=self.client_secret,
                callback_uri=self.callback_url,
                cookie_domain=self.cookie_domain,
                cookie_name=self.cookie_name,
                secure_cookie=self.use_secure_cookie,
                redirect_uri=self.success_url,
            )
        )



import pathlib

import falcon
import falcon.testing
import jwt
import pytest

from falcon_helpers.middlewares.parsejwt import ParseJWTMiddleware as MW
from falcon_helpers.config import ConfigurationError


@pytest.fixture(scope='module')
def rsa_priv_key():
    p = pathlib.Path(__file__).parent.parent.joinpath('files', 'keys', 'testkey')

    with open(p.as_posix(), mode='r') as fd:
        data = fd.read()
    return data


@pytest.fixture(scope='module')
def rsa_pub_key():
    p = pathlib.Path(__file__).parent.parent.joinpath('files', 'keys', 'testkey.pub')

    with open(p.as_posix(), mode='r') as fd:
        data = fd.read()
    return data


@pytest.fixture()
def hs_token():
    return jwt.encode({'key': 'hs', 'aud': 'test'}, 'secret', algorithm='HS256')


@pytest.fixture()
def rsa_token(rsa_priv_key):
    return jwt.encode({'key': 'rsa', 'aud': 'test'}, rsa_priv_key, algorithm='RS256')


def test_init_requires_one_of_a_cookie_or_header():
    with pytest.raises(ConfigurationError):
        MW(audience='blah')

    with pytest.raises(ConfigurationError):
        MW(audience='blah', cookie_name='cookie', header_name='header')

    assert MW(audience='blah', cookie_name='cookie')
    assert MW(audience='blah', header_name='cookie')


def test_init_pubkeys_checks():
    with pytest.raises(ConfigurationError):
        MW(audience='blah', cookie_name='cookie', pubkey='blank')

    assert MW(audience='blah', cookie_name='cookie', pubkey='ssh-rsa blah')
    assert MW(audience='blah', cookie_name='cookie', pubkey=None)


def test_verify_request_fails_without_token():
    mw = MW(audience='blah', cookie_name='cookie', secret='secret')

    with pytest.raises(ValueError):
        mw.verify_request(None)


def test_verification_types(hs_token, rsa_token, rsa_pub_key):
    with pytest.raises(ConfigurationError) as e:
        mw = MW(audience='test', cookie_name='cookie')
        mw.verify_request(hs_token)

    assert 'HS256' in str(e.value)
    assert 'requires a secret key.' in str(e.value)

    mw = MW(audience='test', cookie_name='cookie', secret='secret', decode_algorithms=['HS256'])
    assert mw.verify_request(hs_token) == {
        'key': 'hs',
        'aud': 'test',
    }

    with pytest.raises(ConfigurationError) as e:
        mw = MW(audience='test', cookie_name='cookie')
        mw.verify_request(rsa_token)

    assert 'RS256' in str(e.value)
    assert 'requires a public key.' in str(e.value)

    mw = MW(audience='test', cookie_name='cookie', pubkey=rsa_pub_key)
    assert mw.verify_request(rsa_token) == {
        'key': 'rsa',
        'aud': 'test',
    }


def test_process_request_with_header(hs_token):
    app = falcon.API(
        middleware=[
            MW(audience='test', secret='secret', header_name='Auth', decode_algorithms=['HS256'])
        ]
    )
    client = falcon.testing.TestClient(app)

    resc = falcon.testing.SimpleTestResource()
    app.add_route('/', resc)

    resp = client.simulate_get('/')
    assert resp.status_code == 200
    assert 'auth_token_contents' not in resc.captured_req.context

    resp = client.simulate_get(
        '/',
        headers={
            'Auth': hs_token.decode(),
        }
    )
    assert resp.status_code == 200
    assert 'auth_token_contents' in resc.captured_req.context
    assert resc.captured_req.context['auth_token_contents'] == {
        'key': 'hs',
        'aud': 'test',
    }


def test_process_request_with_cookie(hs_token):
    app = falcon.API(middleware=[
        MW(audience='test', secret='secret', cookie_name='Auth', decode_algorithms=['HS256'])
    ])
    client = falcon.testing.TestClient(app)

    resc = falcon.testing.SimpleTestResource()
    app.add_route('/', resc)

    resp = client.simulate_get('/')
    assert resp.status_code == 200
    assert 'auth_token_contents' not in resc.captured_req.context

    resp = client.simulate_get('/', headers={'Cookie': 'Auth=' + hs_token.decode()})
    assert resp.status_code == 200
    assert 'auth_token_contents' in resc.captured_req.context
    assert resc.captured_req.context['auth_token_contents'] == {
        'key': 'hs',
        'aud': 'test',
    }


def test_process_request_with_default_failed_action():
    app = falcon.API(
        middleware=[MW(audience='test', secret='secret', header_name='Auth')]
    )
    client = falcon.testing.TestClient(app)

    resc = falcon.testing.SimpleTestResource()
    app.add_route('/', resc)

    resp = client.simulate_get('/', headers={'Auth': 'token'})
    assert resp.status_code == 200
    assert 'auth_token_contents' not in resc.captured_req.context


def test_process_request_with_custom_failed_action():
    def custom_failed(exc, req, resp):
        raise RuntimeError("works")

    app = falcon.API(
        middleware=[MW(audience='test', secret='secret', header_name='Auth',
                       when_fails=custom_failed)]
    )
    client = falcon.testing.TestClient(app)

    resc = falcon.testing.SimpleTestResource()
    app.add_route('/', resc)

    with pytest.raises(RuntimeError) as e:
        client.simulate_get('/', headers={'Auth': 'token'})

    assert 'works' == str(e.value)



import datetime
import random
import string

import sqlalchemy as sa
from . import columns


def random_data_for_type(column, numeric_low=-100, numeric_high=100):
    """Return random data based on the data type

    :param column: The SQLA Column to consider


    """
    def randstring(len):
        return "".join([random.choice(string.ascii_letters)
                        for _ in range(len)])

    if isinstance(column.type, sa.types.Enum):
        return random.choice(column.type.enums)
    elif isinstance(column.type, sa.types.Boolean):
        return random.choice([True, False])
    elif isinstance(column.type, sa.types.Integer):
        return random.randint(numeric_high, numeric_low)
    elif isinstance(column.type, sa.types.Numeric):
        return random.uniform(numeric_high, numeric_low)
    elif isinstance(column.type, sa.types.Date):
        return datetime.date.today()
    elif isinstance(column.type, sa.types.DateTime):
        return datetime.datetime.utcnow()
    elif isinstance(column.type, columns.Point):
        return (0, 0)
    elif isinstance(column.type, (sa.types.String, sa.types.Unicode)):
        return randstring(min(column.type.length or 25, 25))

    raise ValueError('No randomization for this column type')



import pytest
import sqlalchemy as sa

from falcon_helpers.app import API
from falcon_helpers.plugins import SQLAlchemyPlugin


def test_plugin_takes_string_url():
    assert SQLAlchemyPlugin('sqlite://').url == sa.engine.url.make_url('sqlite://')


def test_plugin_takes_dict():
    assert SQLAlchemyPlugin({
        'drivername': 'sqlite',
        'host': 'localhost',
        'username': 'myuser',
        'password': 'mypassword',
    }).url == sa.engine.url.make_url('sqlite://myuser:mypassword@localhost')


def test_plugin_takes_url():
    url = sa.engine.url.URL(**{
        'drivername': 'sqlite',
    })
    assert SQLAlchemyPlugin(url).url == sa.engine.url.make_url('sqlite://')


def test_plugin_raises_on_bad_url():
    with pytest.raises(RuntimeError):
        SQLAlchemyPlugin(['bad'])


def test_register_plugin():
    url = sa.engine.url.URL(**{
        'drivername': 'sqlite',
    })
    sess = sa.orm.session.Session()
    pl = SQLAlchemyPlugin(url=url, session=sess)

    app = API()
    pl.register(app)

    assert app.plugins['sqla'] == pl
    assert pl.engine.url == url

    assert pl.session.bind == pl.engine

    assert app._dynmw.req_mw == []
    assert app._dynmw.resc_mw == [pl.mw.process_resource]
    assert app._dynmw.resp_mw == [pl.mw.process_response]


def test_register_plugin_with_sessionmaker():
    url = sa.engine.url.URL(**{
        'drivername': 'sqlite',
    })
    sess = sa.orm.session.sessionmaker()
    pl = SQLAlchemyPlugin(url=url, session=sess)

    app = API()
    pl.register(app)

    assert pl.session().bind == pl.engine



import datetime
import logging
import inspect

import wrapt

log = logging.getLogger(__name__)


def logrequest(level=logging.DEBUG):
    """Wrap resource functions or classes to get logging from called methods


    You can use it on a specific method...

        class MyResource:

            @logrequest(level=logging.INFO)
            def on_get(self, req, resp):
                # ...

    or on an entire class to decorate all falcon resource methods...

        @logrequest(level=logging.INFO)
        class MyResource:

            # Will be decorated
            def on_get(self, req, resp):
                # ...

            # Will be decorated
            def on_post(self, req, resp):
                # will be decorated
    """

    def handle_resource_func(wrapped, instance, args, kwargs):
        start = datetime.datetime.now()
        method_name = f'{instance.__class__.__name__}.{wrapped.__func__.__name__}'
        try:
            log.log(level, f'{method_name} called.')
            result = wrapped(*args, **kwargs)
        except Exception as e:
            end = datetime.datetime.now()
            duration = end - start

            name = e.__class__.__name__
            log.log(level, f'{method_name} raised {name} in {duration.total_seconds()} seconds: {str(e)}.')
            raise
        else:
            end = datetime.datetime.now()
            duration = end - start
            log.log(level, f'{method_name} succeeded in {duration.total_seconds()} seconds.')

        return result

    def handle_resource_class(wrapped, instance, args, kwargs):
        for method in ['__call__', 'on_get', 'on_post', 'on_put', 'on_delete']:
            m = getattr(wrapped, method, None)

            if m:
                setattr(wrapped, method, wrapt.FunctionWrapper(m, handle_resource_func))
        return wrapped(*args, **kwargs)

    @wrapt.decorator
    def logrequest(wrapped, instance, args, kwargs):
        if instance is None:
            if inspect.isclass(wrapped):
                # Decorator was applied to a class.
                return handle_resource_class(wrapped, instance, args, kwargs)
            else:
                # Decorator was applied to a function or staticmethod.
                log.warning(
                    'The logrequest decorator was not attached to an instance method or a '
                    'resource class. This is a noop.')
                return wrapped(*args, **kwargs)
        else:
            if inspect.isclass(instance):
                # Decorator was applied to a classmethod.
                log.warning(
                    'The logrequest decorator was attached to a classmethod or a this is a noop.'
                )
                return wrapped(*args, **kwargs)
            else:
                # Decorator was applied to an instancemethod.
                return handle_resource_func(wrapped, instance, args, kwargs)

    return logrequest


@wrapt.decorator
def timeit(wrapped, instance, args, kwargs):
    method_name = f'{instance.__class__.__name__}.{wrapped.__func__.__name__}'
    start = datetime.datetime.now()

    result = wrapped(*args, **kwargs)

    end = datetime.datetime.now()
    duration = (end - start).total_seconds()
    log.debug(f'{method_name} started at {start} and ended at {end}, total {duration}')

    return result


class Logging:

    def __init__(self, level=logging.INFO, fmt=None):
        self.log_level = level
        self.fmt = fmt or '[%(levelname)s] %(asctime)s - %(msg)s'

    def add_logger(self, name, handler, fmt=None, formatter=None, level=None):
        formatter = (
            logging.Formatter(fmt) if (fmt is not None and not formatter)
            else formatter if formatter
            else logging.Formatter(self.fmt)
        )

        handler.setFormatter(formatter)

        logger = logging.getLogger(name)
        logger.addHandler(handler)
        logger.setLevel(level or self.log_level)



from collections import Iterable

def flatten(items):
    """Yield items from any nested iterable; see REF."""
    for x in items:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            yield from flatten(x)
        else:
            yield x



import falcon_helpers.sqla.db as db


class SQLAlchemySessionMiddleware:
    def __init__(self, session=None):
        self.session = session or db.session

    def process_resource(self, req, resp, resource, params):
        resource.session = self.session

    def process_response(self, req, resp, resource, req_succeeded):
        if not hasattr(resource, 'session'):
            return

        try:
            if not req_succeeded:
                self.session.rollback()
            else:
                self.session.commit()

        except Exception as e:
            self.session.remove()
            raise



import falcon

from falcon_helpers.config import Config
from falcon_helpers.middlewares import multi as multimw


class API(falcon.API):
    __slots__ = (
        'config',
        'plugins',
        '_dynmw',
        'enable_dynamic_mw',
    )

    def __init__(self, middleware=None, enable_dynamic_mw=True, independent_middleware=True,
                 **kwargs):

        if enable_dynamic_mw and not independent_middleware:
            raise RuntimeError(
                f'Independent middleware must be enabled to use dynamic middleware. Either turn '
                'off dynamic middleware (enable_dynamic_mw=False) or enable independent middleware '
                '(independent_middleware=True).'
            )
        elif enable_dynamic_mw:
            self._dynmw = multimw.MultiMiddleware(middleware)
            kwargs['middleware'] = [self._dynmw]
            kwargs['independent_middleware'] = True
        else:
            kwargs['middleware'] = middleware

        self.plugins = {}

        super().__init__(**kwargs)

    def add_middleware(self, mw):
        self._dynmw.add_middleware(mw)

    @classmethod
    def from_inis(cls, *paths, api_kwargs=None):
        """Create an instance of the API from configuration files"""

        app = cls(**(api_kwargs or {}))

        app.config = Config.from_inis(*paths)

        return app



import logging

import pytest

from falcon_helpers.contrib.logging import logrequest


@logrequest()
class FakeResource():
    def on_get(self, req, resp):
        pass

    def on_post(self, req, resp):
        raise Exception('Bad')

    def on_put(self, req, resp):
        pass

    def on_delete(self, req, resp):
        pass


class FakeResource2():
    @logrequest()
    def on_get(self, req, resp):
        pass


def test_class_decoration(client, caplog):

    client.app.add_route('/', FakeResource())

    with caplog.at_level(logging.DEBUG):
        client.simulate_get('/')

        assert len(caplog.records) == 2
        start, finish = caplog.records
        assert start.message.startswith('FakeResource.on_get called')
        assert finish.message.startswith('FakeResource.on_get succeeded')

    caplog.clear()

    with caplog.at_level(logging.DEBUG):
        with pytest.raises(Exception):
            client.simulate_post('/')

        assert len(caplog.records) == 2
        start, finish = caplog.records
        assert start.message.startswith('FakeResource.on_post called')
        assert finish.message.startswith('FakeResource.on_post raised Exception')


def test_method_decoration(client, caplog):

    client.app.add_route('/', FakeResource2())

    with caplog.at_level(logging.DEBUG):
        client.simulate_get('/')

        assert len(caplog.records) == 2
        start, finish = caplog.records
        assert start.message.startswith('FakeResource2.on_get called')
        assert finish.message.startswith('FakeResource2.on_get succeeded')



from datetime import timezone as tz
import pytest
import sqlalchemy as sa

import falcon
import falcon.testing
import marshmallow_sqlalchemy as ms

from falcon_helpers.resources.crud import CrudBase, ListBase
from falcon_helpers.middlewares.sqla import SQLAlchemySessionMiddleware
from falcon_helpers.middlewares.marshmallow import MarshmallowMiddleware
from falcon_helpers.sqla.orm import BaseColumns, BaseFunctions, Testable
from falcon_helpers.sqla.db import session

from falcon_helpers.tests.fixtures import Base


class ModelOther(Base, BaseColumns, Testable):
    __tablename__ = 'mother'

    test_id = sa.Column(sa.Integer, sa.ForeignKey('mtest.id'), nullable=False)


class ModelTest(Base, BaseColumns, BaseFunctions, Testable):
    __tablename__ = 'mtest'

    name = sa.Column(sa.Unicode, nullable=False)
    uni = sa.Column(sa.Unicode, unique=True)
    other = sa.orm.relationship("ModelOther")


class ModelSchema(ms.ModelSchema):
    class Meta:
        model = ModelTest
        exclude = ('other',)


class BasicCrud(CrudBase):
    db_cls = ModelTest
    schema = ModelSchema

class List(ListBase):
    db_cls = ModelTest
    schema = ModelSchema

class ListSub(ListBase):
    db_cls = ModelTest
    schema = ModelSchema

    def get_objects(self, req, *args, **kwargs):
        if kwargs['objid'] == 'missing':
            return None

        if kwargs['objid'] == 'zero':
            return []


@pytest.fixture
def app():
    Base.metadata.drop_all()
    Base.metadata.create_all()

    app = falcon.API(
        middleware=[
            SQLAlchemySessionMiddleware(session),
            MarshmallowMiddleware(),
        ]
    )

    app.add_route('/crud/{obj_id}', BasicCrud())
    app.add_route('/bad', BasicCrud())
    app.add_route('/list', List())
    app.add_route('/list/{objid}/other', ListSub())
    return app


@pytest.fixture
def client(app):
    return falcon.testing.TestClient(app)


class TestCrudBase:
    def test_crud_base_get_404_with_no_object(self, client):
        resp = client.simulate_get('/crud/1')
        assert resp.status_code == 404

    def test_crud_base_get_500_with_misconfigured_route(self, client):
        resp = client.simulate_get('/bad')
        assert resp.status_code == 500

    def test_crud_base_get_200_with_object(self, client):
        m1 = ModelTest.testing_create()
        resp = client.simulate_get(f'/crud/{m1.id}')

        assert resp.status_code == 200
        assert resp.json == {
            'id': m1.id,
            'name': m1.name,
            'uni': m1.uni,
            'created_ts': m1.created_ts.replace(tzinfo=tz.utc).isoformat(),
            'updated_ts': m1.updated_ts.replace(tzinfo=tz.utc).isoformat(),
        }

    def test_crud_base_get_404_with_bad_primary_key(self, client):
        assert client.simulate_get(f'/crud/abs').status_code == 404

    def test_crud_base_post_duplicate_object(self, client):
        ModelTest.testing_create(uni='test')
        resp = client.simulate_post(
            f'/crud/new',
            json={
                'uni': 'test',
                'name': 'thing'
            })

        assert resp.status_code == 409

    def test_crud_base_get_404_with_wrong_id(self, client):
        m1 = ModelTest.testing_create()
        session.add(m1)
        session.commit()
        resp = client.simulate_get(f'/crud/{m1.id + 1}')

        assert resp.status_code == 404

    def test_crud_base_post(self, client):
        resp = client.simulate_post(
            f'/crud/new',
            json={
                'name': 'thing'
            })

        assert resp.status_code == 201
        assert session.query(ModelTest).get(resp.json['id']).name == 'thing'
        assert resp.json['name'] == 'thing'

    def test_crud_base_delete(self, client):
        m1 = ModelTest.testing_create()
        session.add(m1)
        session.commit()

        resp = client.simulate_delete(f'/crud/{m1.id}')

        assert resp.status_code == 204
        assert session.query(ModelTest).get(m1.id) is None

    def test_crud_base_delete_with_relationship(self, client):
        m1 = ModelTest.testing_create()
        session.add(m1)
        session.flush()

        mo1 = ModelOther.testing_create(test_id=m1.id)
        session.add(mo1)
        session.commit()

        resp = client.simulate_delete(f'/crud/{m1.id}')

        assert resp.status_code == 400
        assert session.query(ModelTest).get(m1.id) == m1
        assert session.query(ModelOther).get(mo1.id) == mo1

        assert 'errors' in resp.json
        assert resp.json['errors'] == [
            'Unable to delete because the object is connected to other objects'
        ]


class TestListBase:

    def test_default_filter_for_type(self):
        lb = ListSub()
        result = lb.filter_for_column(ModelTest.name, 'name')
        assert result.left == ModelTest.name
        assert result.right.value == 'name'
        assert result.operator.__name__ == 'contains_op'

    def test_default_filter_for_column(self):
        lb = ListSub()
        lb.column_filters = {
            'name': ModelTest.name.__eq__
        }

        result = lb.filter_for_param('name', 'val')
        assert result.left == ModelTest.name
        assert result.right.value == 'val'
        assert result.operator.__name__ == 'eq'

    def test_column_type_filter_creations(self):
        lb = ListSub()
        lb.type_filters = {
            (sa.sql.sqltypes.String, lambda a, b: 'works')
        }

        # check the defaults
        assert sa.sql.sqltypes.String in lb.column_type_filters

        # check the overriding
        assert lb.column_type_filters[sa.sql.sqltypes.String]('col', 'val') == 'works'

    def test_listbase_get_with_default_pagination(self, app):
        m1 = ModelTest.testing_create()
        m2 = ModelTest.testing_create()
        session.add_all([m1, m2])
        session.commit()

        resource = List()
        resource.default_page_size = None

        app.add_route('/pagination', resource)
        c = client(app)

        resp = c.simulate_get(f'/pagination')
        assert len(resp.json) == 2

        resp = c.simulate_get(f'/pagination', query_string=f'pageSize=1')
        assert len(resp.json) == 1

    def test_listbase_get(self, client):
        m1 = ModelTest.testing_create()
        m2 = ModelTest.testing_create()
        session.add_all([m1, m2])
        session.commit()

        resp = client.simulate_get(f'/list', query_string=f'name={m1.name}')

        assert len(resp.json) == 1
        assert resp.json[0]['id'] == m1.id
        assert resp.json[0]['name'] == m1.name

    def test_listbase_get_sends_404_for_subobj_with_none_respose(self, client):
        resp = client.simulate_get(f'/list/missing/other')
        assert resp.status_code == 404
        assert 'error' in resp.json

    def test_listbase_get_sends_200_for_subobj_with_empty_respose(self, client):
        ModelTest.testing_create()

        resp = client.simulate_get(f'/list/zero/other')
        assert resp.status_code == 200
        assert resp.json == []



import configparser


class ConfigurationError(Exception):
    pass


class Config(dict):
    """A basic Configuration type

    This configuration class can use subscriptable syntax `['item']` or dotted syntax `config.item`
    """

    def __getattr__(self, item):
        return self.__getitem__(item)

    def __setattr__(self, item, value):
        return self.__setitem__(item, value)

    def __delattr__(self, item):
        return self.__delitem__(item)

    def __getitem__(self, item):
        try:
            return super().__getitem__(item)
        except KeyError as e:
            raise ConfigurationError(f'Unable to find configuration value for key {item}')

    @classmethod
    def from_inis(cls, *fpaths, coerce_bools=True):
        """Read the configuration for ini files

        *fpaths is passed straight to `ConfigParser.read` so it can be multiple configuration files
        and it will read them in order.
        """
        config = configparser.ConfigParser()
        config.read(fpaths)

        sections = []
        for section in config.sections():
            items = []
            for key, value in config.items(section):
                # Coerce booleans to literal types
                if value == 'true' or value == 'True':
                    value = True

                elif value == 'false' or value == 'False':
                    value = False

                items.append((key, value))

            sections.append((section, cls(items)))
        return cls(sections)



import logging
import ujson
import sqlalchemy as sa
import falcon

from ..utils import flatten


log = logging.getLogger(__name__)


class ListBase:
    """A base class for returning list of objects.

    This base class assumes you are using the marshmallow middleware for object
    serialization and sqlalchemy for database access.

    Attributes:
        db_cls: Set this to a SQLAlchemy entity
        schema: Set this a Marshmallow schema

    Filtering:

        Filtering allows you to specify custom logic to limit query results while still using the
        standard ListBase query builder.

        It is possible to define custom type and custom column filters on names properties by using
        the type_filters and the column_filters settings appropriately. `column_filters` will be
        used first and then fallback to a `type_filter`. Basic type filters are defined for common
        types where a sensible default is chosen. Note that `Integer` and `Float` will try to do the
        conversion and error if bad data is sent.


        To specify column filters, assign the `column_filters` class variable to a dictionary
        containing the name of the key and the function(value) you want to call.

            class MyEntityList(ListBase):
                db_cls = MyEntity
                schema = MyEntitySchema

                column_filters = {
                    'my_column': MyEntity.filters_for_string
                }

        To add additional type filters use the `type_filters` class variable.

            class MyEntityList(ListBase):
                db_cls = MyEntity
                schema = MyEntitySchema

                type_filters = {
                    sqlalchemy.types.DateTime: MyEntity.filters_for_string
                }


    """
    db_cls = None
    schema = None

    default_order = None

    # TODO: Deprecate the 50 page size limit and make this None
    default_page_size = 50

    _default_type_filters = frozenset([
        (sa.sql.sqltypes.String,  lambda col, val: col.contains(val)),
        (sa.sql.sqltypes.Unicode, lambda col, val: col.contains(val)),
        (sa.sql.sqltypes.Integer, lambda col, val: col == int(val)),
        (sa.sql.sqltypes.Numeric, lambda col, val: col == float(val)),
    ])
    type_filters = None

    column_filters = None

    @property
    def column_type_filters(self):
        filters = dict()
        filters.update({k: v for k, v in self._default_type_filters})
        filters.update(self.type_filters or {})
        return filters

    def on_get(self, req, resp, **kwargs):
        result = self.get_objects(req, **kwargs)

        if result is None:
            resp.status = falcon.HTTP_404
            resp.media = {'error': 'Unable to find objects with that identifier'}
            return

        resp.status = falcon.HTTP_200
        resp.body = self.response_data(result, req, **kwargs)

    def response_data(self, data, req, **kwargs):
        schema = self.schema(many=True)
        return schema.dump(data)

    def base_query(self, req, **kwargs):
        return self.session.query(self.db_cls)

    def pagination_hook(self, query, req, **kwargs):
        """Create a hook for pagination"""
        size = req.params.get('pageSize')

        if not size:
            size = self.default_page_size
        else:
            size = int(size)

        # -1 here is so that the page numbers start at 1
        page = int(req.params.get('page', 1)) - 1

        if page < 0:
            page = 0

        if size:
            return query.limit(size).offset((page * size))
        else:
            return query

    def columns_for_params(self, params):
        return {getattr(self.db_cls, x)
                for x in self.db_cls.orm_column_names() & params.keys()}

    def filter_for_column(self, column, value):
        filters = self.column_type_filters
        if (type(column.property) == sa.orm.properties.ColumnProperty
                and type(column.type) in filters):
            return filters[type(column.type)](column, value)

        return None

    def filter_for_param(self, key, value):
        try:
            f = self.column_filters[key]
        except KeyError:
            return None

        return f(value)

    def filter_hook(self, query, req, **kwargs):
        filters = {col.key: self.filter_for_column(col, req.params[col.key])
                   for col in self.columns_for_params(req.params)}

        # Custom column filters
        if self.column_filters:
            filters.update({key: self.filter_for_param(key, value)
                            for key, value in req.params.items()})
        return query.filter(*[x for x in filters.values() if x is not None])

    def order_hook(self, query, req, **kwargs):
        request_order = req.params.get('sort_by', [])

        # This takes advantage of falcons duplicate-param parsing which returns
        # a list when it finds a param multiple times.
        if type(request_order) != list:
            request_order = [request_order]

        request_order = list(flatten([x.split(';') for x in request_order]))
        request_order = [sa.desc(x[1:]) if x.startswith('-') else sa.asc(x) for x in request_order]

        default_order = request_order or self.default_order or self.db_cls.__mapper__.primary_key
        return query.order_by(*default_order)

    def custom_hook(self, query, req, **kwargs):
        return query

    def get_objects(self, req, *args, **kwargs):
        base = self.base_query(req, **kwargs)
        filtered = self.filter_hook(base, req, **kwargs)
        order = self.order_hook(filtered, req, **kwargs)
        paged = self.pagination_hook(order, req, **kwargs)
        final = self.custom_hook(paged, req, **kwargs)

        return final.all()


class CrudBase:
    """A very simple CRUD resource.

    This base class assumes you are using the marshmallow middleware for object
    serialization and sqlalchemy for database access.

    Attributes:
        db_cls: Set this to a SQLAlchemy entity
        schema: Set this a Marshmallow schema
    """
    db_cls = None
    schema = None
    default_param_name = 'obj_id'

    def delete_object(self, obj, req, **kwargs):
        self.session.delete(obj)
        self.session.flush()

    def get_object(self, req, **kwargs):
        try:
            obj_id = kwargs[self.default_param_name]
        except KeyError:
            log.error(
                f'The resource {self.__class__.__name__} route is not using the correct parameter '
                f'name for the object identifier. Expecting `{self.default_param_name}` but it was '
                f'not in the matched route parameters. Add a `default_param_name` to the resource '
                f'which matches the route variable. Found these items: {",".join(kwargs.keys())}.'
            )
            raise falcon.HTTPInternalServerError("Misconfigured route")

        try:
            return self.session.query(self.db_cls).get(obj_id)
        except sa.exc.DataError as e:
            self.session.rollback()
            log.warning(f'Bad primary key given to  {self.__class__.__name__}')
            return None

    def on_get(self, req, resp, **kwargs):
        result = self.get_object(req, **kwargs)

        if not result:
            raise falcon.HTTPNotFound

        schema = self.schema()
        resp.body = schema.dump(result)
        resp.status = falcon.HTTP_200

    def on_put(self, req, resp, **kwargs):
        self.session.add(req.context['dto'].data)
        self.session.flush()

        resp.status = falcon.HTTP_200
        resp.body = self.schema().dump(req.context['dto'].data)

    def on_post(self, req, resp, **kwargs):
        self.session.add(req.context['dto'].data)

        try:
            self.session.flush()
        except sa.exc.IntegrityError as e:
            self.session.rollback()

            if 'violates not-null constraint' in str(e):
                resp.status = falcon.HTTP_400
                resp.media = {
                    'errors': ['The submitted object is not complete.']
                }
            else:
                resp.status = falcon.HTTP_409
                resp.media = {
                    'errors': ['An object with that identifier already exists.']
                }
            return

        resp.status = falcon.HTTP_201
        resp.media = self.schema().dump(req.context['dto'].data).data

    def on_delete(self, req, resp, **kwargs):
        try:
            obj = self.get_object(req, **kwargs)

            if obj:
                self.delete_object(obj, req, **kwargs)

        except sa.exc.IntegrityError as e:
            self.session.rollback()
            resp.status = falcon.HTTP_400
            resp.media = {'errors': [('Unable to delete because the object is '
                                      'connected to other objects')]}
        else:
            resp.status = falcon.HTTP_204



import falcon.testing
import pytest
import sqlalchemy as sa

from falcon_helpers.middlewares.sqla import SQLAlchemySessionMiddleware
import falcon_helpers.sqla.db as db


class MockResource(falcon.testing.SimpleTestResource):

    def on_delete(self, req, resp):
        self.session.execute('BAD')
        self.session.execute('BAD')


@pytest.fixture()
def client():
    api = falcon.API(middleware=[
        SQLAlchemySessionMiddleware()
    ])

    return falcon.testing.TestClient(api)


def test_mw_takes_custom_session(client):
    sess = sa.orm.session.Session()
    assert db.session != sess

    mw = SQLAlchemySessionMiddleware(sess)
    api = falcon.API(middleware=[mw])
    resc = MockResource()
    api.add_route('/', resc)
    custom_client = falcon.testing.TestClient(api)

    custom_client.simulate_get('/')
    assert resc.session == sess

    resc2 = MockResource()
    client.app.add_route('/', resc2)
    client.simulate_get('/')
    assert resc2.session == db.session



import falcon.testing

from falcon_helpers.middlewares.multi import MultiMiddleware


class FakeMW:
    def process_request(self, req, resp):
        req.context['req'] = True

    def process_resource(self, req, resp, resource, params):
        req.context['resc'] = True

    def process_response(self, req, resp, resource, resp_succeded):
        media = resp.media
        media['resp'] = True
        resp.media = media


class FakeResc:
    def on_get(self, req, resp):
        resp.media = req.context


def test_accepts_middleware_during_init():
    mw = MultiMiddleware()

    assert mw.req_mw == []
    assert mw.resc_mw == []
    assert mw.resp_mw == []

    mw = MultiMiddleware(middleware=FakeMW())
    assert len(mw.req_mw) == 1
    assert len(mw.resc_mw) == 1
    assert len(mw.resp_mw) == 1

    mw = MultiMiddleware(middleware=[FakeMW(), FakeMW()])
    assert len(mw.req_mw) == 2
    assert len(mw.resc_mw) == 2
    assert len(mw.resp_mw) == 2


def test_calls_middlewares():
    app = falcon.API(middleware=MultiMiddleware(middleware=FakeMW()))

    app.add_route('/', FakeResc())

    c = falcon.testing.TestClient(app)
    resp = c.simulate_get('/')

    assert resp.json == {
        'req': True,
        'resc': True,
        'resp': True,
    }



import falcon.testing
import pytest

from falcon_helpers import API
from falcon_helpers.plugins import SentryPlugin


class FakeException(Exception):
    pass


@pytest.fixture()
def app():
    app = API()

    class FakeResource:
        def on_get(self, req, resp):
            raise FakeException('Failed')

    app.add_route('/fails', FakeResource())

    return app


@pytest.fixture()
def client(app):
    return falcon.testing.TestClient(app)


def test_sentry_with_dsn(mocked_sentry_client, client):
    plugin = SentryPlugin('test_dsn')
    plugin.register(client.app)

    client.simulate_get('/fails')
    name, args, kwargs = mocked_sentry_client.method_calls[0]

    assert plugin.dsn == 'test_dsn'
    assert isinstance(args[0], FakeException)


def test_sentry_without_dsn_still_raise(mocked_sentry_client, client):
    plugin = SentryPlugin()
    plugin.register(client.app)

    with pytest.raises(FakeException):
        client.simulate_get('/fails')


def test_sentry_pulls_from_app_config(mocked_sentry_client, client):
    plugin = SentryPlugin()
    client.app.config = {
        'sentry': {
            'dsn': 'something',
            'environment': 'prod',
        }
    }
    plugin.register(client.app)

    assert plugin.dsn == 'something'
    assert plugin.environment == 'prod'



import pytest
import sqlalchemy as sa
import unittest.mock

import falcon.testing
import falcon_helpers.app as app
import falcon_helpers.sqla.db as db

bind = sa.engine.create_engine('sqlite://')
Base = sa.ext.declarative.declarative_base(bind=bind)

db.session.configure(bind=bind)


@pytest.fixture()
def api():
    return app.API()


@pytest.fixture()
def client(api):
    return falcon.testing.TestClient(api)


@pytest.fixture()
def mocked_sentry_client():
    with unittest.mock.patch(
        'falcon_helpers.plugins.sentry.raven.Client',
        spec_set=True,
        autospec=True
    ) as m:

        yield m



import logging
import wsgiref.simple_server


log = logging.getLogger(__name__)


class CustomLoggingWSGIRequestHandler(wsgiref.simple_server.WSGIRequestHandler):

    def log_message(self, format, *args):
        log.info(f'{format%args}')



import boto3
import moto
import os.path
import pathlib
import tempfile
import falcon_helpers.contrib.storage as storage


def document_from_temp(temp):
    return storage.Document(
        name=temp.name,
        uid=os.path.basename(temp.name),
        storage_type='local',
        path=temp.name,
        details={}
    )


class TestS3FileStore:
    @moto.mock_s3
    def test_warning_message_writes_when_passing_mode(self, caplog):
        s3 = boto3.resource('s3')
        s3.create_bucket(Bucket='bucket')
        obj = s3.Object('bucket', key='path')
        obj.put(Body='content')

        doc = storage.Document('name', 'uid', 'path')
        store = storage.S3FileStore('bucket')
        data = store.fetch_fp(doc, mode='rb')
        assert data.read() == b'content'
        assert 'S3FileStore' in caplog.records[0].message


class TestLocalFileStore:
    def test_fetching_a_file_by_fp(self):
        with tempfile.TemporaryDirectory() as d:
            store = storage.LocalFileStore(d, uidgen=lambda: 'unique-name')

            with tempfile.NamedTemporaryFile(prefix=d) as rf:
                rf.write(b'test')
                rf.seek(0)

                doc = document_from_temp(rf)
                fp = store.fetch_fp(doc)
                with fp as f:
                    assert f.read() == b'test'

    def test_save_does_not_require_path(self):
        with tempfile.TemporaryDirectory() as d:
            store = storage.LocalFileStore(d, uidgen=lambda: 'unique-name')

            with tempfile.NamedTemporaryFile(prefix=d) as f:
                doc = store.save(f.name, f)
                assert doc.uid == 'unique-name'
                assert doc.name == f.name
                assert doc.path == str(pathlib.Path(d).joinpath(doc.uid))

            with tempfile.NamedTemporaryFile(prefix=d) as f:
                doc = store.save(f.name, f, path='other')
                assert doc.uid == 'unique-name'
                assert doc.name == f.name
                assert doc.path == str(pathlib.Path(d).joinpath('other', doc.uid))



from .middlewares.jinja2 import Jinja2Middleware
from .middlewares.auth_required import AuthRequiredMiddleware

from .resources import auth0
from .config import Config
from .app import API

from . import contrib



import sqlalchemy as sa

import falcon_helpers.sqla.db as db


class LoadUserMiddleware:
    """Load a user from the database during the request cycle using sqlalchemy

    By default this will grab the id data from the auth token on the request
    context assuming you are using `.middleware.auth_required`. To change that
    pass a function which takes the request and returns the user id to `get_id`.

    :param session: a sqlalchemy session
    :param user_cls: a class which will return the user object. This object must
        have a `get_by_id` which will return a SQLAlchemy Query.
    :param get_id: a function which will get the user identifier off of the
        request.

    """

    def __init__(self, user_cls, get_id=None, session=None):
        self.user_cls = user_cls
        self.get_id = get_id or self._get_id
        self.session = session or db.session

    @staticmethod
    def _get_id(req):
        try:
            return req.context.get('auth_token_contents').get('sub')
        except AttributeError as e:
            return None

    def fetch_user(self, user_id):
        return (self.user_cls
                    .get_by_id(user_id)
                    .with_session(self.session())
                    .one_or_none())

    def process_request(self, req, resp):
        user_id = self.get_id(req)
        req.context['user'] = self.fetch_user(user_id)



from .jinja2 import (
    Jinja2ConfigurationError,
    Jinja2Middleware,
    Jinja2Response,
)
from .sqla import SQLAlchemySessionMiddleware

from .auth_required import AuthRequiredMiddleware

from .marshmallow import MarshmallowMiddleware

from .load_user import LoadUserMiddleware
from .parsejwt import ParseJWTMiddleware
from .multipart import MultipartMiddleware



import logging
import falcon
import raven

from falcon_helpers.config import ConfigurationError

log = logging.getLogger(__name__)


class SentryPlugin:
    """Simple Sentry integration plugin

        sentry = falcon_helpers.plugins.SentryPlugin()
        app = falcon_helpers.API()
        app.config = {'sentry': {dsn': 'rand_sentry_dsn'}}

        sentry.register_app(app)

    You can get access to this globally so that you can capture exceptions without it having to get
    all the way up to the Falcon layer again.  Any exception that is caught by the falcon error
    handler will be sent to sentry (except HTTPError's and HTTPStatus's).

        sentry = Sentry()

        def create_app(config):
            app = falcon_helpers.API()
            app.config = config

            sentry.register(app)
            return app


        class SomeResource:

            def on_get(self, req, resp):
                try:
                    some_failure()
                except Exception as e:
                    sentry.captureException(e)
                    pass

    You have full access to the client with `sentry_plugin.client` and can update that do have
    special configurations.

    NOTE(nZac): if environment is passed or in the configuration, we will try to set that as well on
    the client to get environment support. By default the environment is `None`.
    """

    __slots__ = (
        '_dsn',
        'client',
        'environment',
    )

    @property
    def dsn(self):
        return self._dsn

    @dsn.setter
    def dsn(self, dsn):
        if dsn is None:
            self._dsn = None
            return

        self._dsn = dsn

        if self.client:
            self.client.set_dsn(self.dsn)

    def __init__(self, dsn=None, environment=None):
        self.client = None
        self.environment = environment
        self.dsn = dsn

    def _make_client(self):
        if not self.dsn:
            log.warning('No sentry client configured, errors will not be reported with Sentry.')

        return raven.Client(
            dsn=self.dsn,
            environment=self.environment,
        )

    def update_settings_from_app_config(self, app):
        """A hook allowing setup of sentry through a config convention.

        Subclass and override as necessary.
        """

        config = getattr(app, 'config', None)

        if config is None:
            return

        if self.environment is None:
            try:
                self.environment = config['sentry']['environment']
            except (KeyError, ConfigurationError):
                self.environment = None

        if self.dsn is None:
            try:
                self.dsn = config['sentry']['dsn']
            except (KeyError, ConfigurationError):
                self.dsn = None

        if not self.client:
            self.client = self._make_client()

    def register(self, app):
        self.update_settings_from_app_config(app)

        if not self.client:
            self.client = self._make_client()

        if self.client:
            app.add_error_handler(Exception, self.handle)

        app.plugins['sentry'] = self

    def handle(self, ex, req, resp, params):
        raisable = (falcon.http_error.HTTPError, falcon.http_status.HTTPStatus)

        if isinstance(ex, raisable):
            raise ex
        elif self.dsn:
            self.client.captureException(ex)
            raise falcon.HTTPInternalServerError()
        else:
            raise



import falcon.testing
import pytest
import sqlalchemy as sa
import marshmallow as mm
import marshmallow_sqlalchemy as mms

import falcon_helpers.sqla.orm as orm
import falcon_helpers.sqla.db as db
from falcon_helpers.middlewares.marshmallow import MarshmallowMiddleware


@pytest.fixture()
def client():
    api = falcon.API(middleware=[
        MarshmallowMiddleware()
    ])

    return falcon.testing.TestClient(api)


class ObjEntity(orm.BaseColumns, orm.ModelBase):
    __tablename__ = 'obj_entity'

    name = sa.Column(sa.String)


class Obj(mms.ModelSchema):
    name = mm.fields.String()

    class Meta:
        model = ObjEntity


class WithoutSchemaResc(falcon.testing.SimpleTestResource):
    pass


class WithSchemaResc(falcon.testing.SimpleTestResource):
    schema = Obj


class WithCustomLoader(WithSchemaResc):
    def schema_loader(self, data, req, resource, params):
        result = self.schema().load(data, session=db.session)
        result.data.name = 'other'
        return result


def test_get_is_not_marshalled(client):
    resource = falcon.testing.SimpleTestResource()

    client.app.add_route('/test', resource)
    resp = client.simulate_get('/test')

    assert resp.status_code == 200
    assert not resource.captured_req.context['_marshalled']


def test_post_is_not_marshalled_without_body(client):
    resource = falcon.testing.SimpleTestResource()

    client.app.add_route('/test', resource)
    resp = client.simulate_post('/test')

    assert resp.status_code == 200
    assert not resource.captured_req.context['_marshalled']


def test_post_is_not_marshalled_with_json_but_has_schema(client):
    resource = WithoutSchemaResc()

    client.app.add_route('/test', resource)
    resp = client.simulate_post(
        '/test',
        json={}
    )

    assert resp.status_code == 200
    assert not resource.captured_req.context['_marshalled']


def test_turning_off_auto_marshalling(client):
    resource = WithSchemaResc()
    resource.auto_marshall = False

    client.app.add_route('/test', resource)
    resp = client.simulate_post(
        '/test',
        json={}
    )

    assert resp.status_code == 200
    assert not resource.captured_req.context['_marshalled']


def test_turning_verify_content_type_is_json(client):
    resource = WithSchemaResc()

    client.app.add_route('/test', resource)
    resp = client.simulate_post(
        '/test',
        body='["looks like json"]'
    )

    assert resp.status_code == 200
    assert not resource.captured_req.context['_marshalled']


def test_turning_verify_content_length(client):
    resource = WithSchemaResc()

    client.app.add_route('/test', resource)
    resp = client.simulate_post(
        '/test',
        headers={'Content-Type': 'application/json'},
        body=''
    )

    assert resp.status_code == 200
    assert not resource.captured_req.context['_marshalled']


def test_the_happy_path(client):
    resource = WithSchemaResc()
    client.app.add_route('/test', resource)
    resp = client.simulate_post(
        '/test',
        json={'name': 'john'}
    )

    assert resp.status_code == 200
    assert resource.captured_req.context['_marshalled']


def test_keeps_the_media_and_populates_the_raw_stream(client):
    resource = WithSchemaResc()
    client.app.add_route('/test', resource)

    resp = client.simulate_post(
        '/test',
        json={'name': 'john'}
    )

    assert resp.status_code == 200
    assert resource.captured_req.context['_marshalled']
    assert resource.captured_req.media == {'name': 'john'}
    assert resource.captured_req.context['marshalled_stream'] == b'{"name":"john"}'


def test_support_default_loader(client):
    resource = WithSchemaResc()
    client.app.add_route('/test', resource)

    resp = client.simulate_post(
        '/test',
        json={'name': 'john'}
    )

    assert resp.status_code == 200
    assert resource.captured_req.context['_marshalled']
    assert resource.captured_req.context['dto'].data.name == 'john'


def test_support_custom_loader(client):
    resource = WithCustomLoader()
    client.app.add_route('/test', resource)

    resp = client.simulate_post(
        '/test',
        json={'name': 'john'}
    )

    assert resp.status_code == 200
    assert resource.captured_req.context['_marshalled']
    assert resource.captured_req.context['dto'].data.name == 'other'


def test_errors_during_loading(client):
    resource = WithSchemaResc()
    client.app.add_route('/test', resource)

    resp = client.simulate_post(
        '/test',
        json={'name': 1}
    )

    assert resp.status_code == 400
    assert resp.json == {'errors': {'name': ['Not a valid string.']}}
    assert resource.captured_req is None



import unittest.mock as mock

import pytest
import click.testing

import falcon_helpers.contrib.cli as cli
import falcon_helpers.contrib.wsgi as wsgi


@pytest.fixture(scope='function')
def cli_with_app():
    @click.group()
    @click.pass_context
    def test(ctx):
        ctx.meta['app'] = 'fake'

    test.add_command(cli.server)
    return test


@pytest.fixture(scope='function')
def runner(request):
    return click.testing.CliRunner()


def test_server_without_host_and_port(runner, cli_with_app):
    with mock.patch('falcon_helpers.contrib.cli.make_server') as m_make_server:
        runner.invoke(cli_with_app, ['server', 'start'])
        m_make_server.assert_called_with(
            '127.0.0.1', 9090, 'fake',
            handler_class=wsgi.CustomLoggingWSGIRequestHandler)


def test_server_with_host(runner, cli_with_app):
    with mock.patch('falcon_helpers.contrib.cli.make_server') as m_make_server:
        runner.invoke(cli_with_app, ['server', 'start', '-h', 'localhost'])
        m_make_server.assert_called_with(
            'localhost', 9090, 'fake',
            handler_class=wsgi.CustomLoggingWSGIRequestHandler)


def test_server_with_port(runner, cli_with_app):
    with mock.patch('falcon_helpers.contrib.cli.make_server') as m_make_server:
        runner.invoke(cli_with_app, ['server', 'start', '-p', '9'])
        m_make_server.assert_called_with(
            '127.0.0.1', 9, 'fake',
            handler_class=wsgi.CustomLoggingWSGIRequestHandler)


def test_server_with_long(runner, cli_with_app):
    with mock.patch('falcon_helpers.contrib.cli.make_server') as m_make_server:
        runner.invoke(cli_with_app, ['server', 'start', '--port', '9', '--host', 'localhost'])
        m_make_server.assert_called_with(
            'localhost', 9, 'fake',
            handler_class=wsgi.CustomLoggingWSGIRequestHandler)


def test_server_without_app(runner):
    with mock.patch('falcon_helpers.contrib.cli.make_server') as m_make_server:
        res = runner.invoke(cli.server, ['start'])
        m_make_server.assert_not_called()

        assert 'Unable to find an application' in res.output



import textwrap
import tempfile
import pytest

from falcon_helpers.config import (
    Config,
    ConfigurationError
)


def test_config_as_dict():
    conf = Config()
    conf['thing'] = 1
    conf['other'] = {'thing': 2}

    assert conf['thing'] == 1
    assert conf['other'] == {'thing': 2}

    with pytest.raises(ConfigurationError):
        assert conf['doesnt']


def test_config_as_mapping():
    conf = Config()
    conf.thing = 1
    conf.other = 'best'

    assert conf.thing == 1
    assert conf.other == 'best'

    with pytest.raises(ConfigurationError):
        assert conf.doesnt


def test_config_from_inis():

    with tempfile.NamedTemporaryFile() as f1, tempfile.NamedTemporaryFile() as f2:
        conf1 = textwrap.dedent(
            '''
            [sectionone]
            x = 1

            [sectiontwo]
            x = 2
            y = somestring
            ''')

        conf2 = textwrap.dedent(
            '''
            [sectionone]
            x = 3
            thing = false
            other = true

            [sectionthree]
            y = 6
            other = mystring
            ''')

        f1.write(conf1.encode())
        f1.seek(0)

        f2.write(conf2.encode())
        f2.seek(0)

        config = Config.from_inis(f1.name, f2.name)

        # Test later config override earlier ones
        assert config.sectionone.x == '3'
        assert config.sectionone.thing is False
        assert config.sectionone.other is True

        # test same named, in different section, is not overridden
        assert config.sectiontwo.x == '2'

        # test early configs without override stick around
        assert config.sectiontwo.y == 'somestring'

        # test later configs with new names are added
        assert config.sectionthree.other == 'mystring'

        # test later configs with names in other sections are added
        assert config.sectionthree.y == '6'



import logging
import falcon
import ujson
from marshmallow.schema import MarshalResult

import falcon_helpers.sqla.db as db

log = logging.getLogger(__name__)


class MarshmallowMiddleware:

    def _default_load(self, data, req, resource, params):
        schema = resource.schema()

        # Try to get an instance from the `get_object` method on the resource so we can populate
        # already existing instances
        if hasattr(resource, 'get_object'):
            instance = resource.get_object(req=req, **params)
        else:
            instance = None

        return schema.load(data, session=db.session, instance=instance)

    def process_resource(self, req, resp, resource, params):
        should_parse = (
            # Veriy that it is safe to parse this resource
            req.method in ('POST', 'PUT'),
            # If there is no data in the body, there is nothing to look at
            bool(req.content_length),
            # If the resource doesn't have a schema the loading would be impossible
            hasattr(resource, 'schema'),
            # If the resource has turned off auto marshaling
            getattr(resource, 'auto_marshall', True),
            # Only consider JSON requests for auto-parsing
            (req.content_type and req.content_type.startswith('application/json')),
        )

        # Check that all the conditions for parsing are met
        if not all(should_parse):
            req.context['_marshalled'] = False
            return

        req.context['marshalled_stream'] = req.stream.read()
        data = req._media = ujson.loads(req.context['marshalled_stream'])

        loaded = (self._default_load(data, req, resource, params)
                  if not hasattr(resource, 'schema_loader')
                  else resource.schema_loader(data, req, resource, params))

        if loaded.errors:
            #  This should probably return whatever the accept header indicates
            raise falcon.HTTPStatus(
                falcon.HTTP_400,
                headers={'Content-Type': 'application/json'},
                body=ujson.dumps({'errors': loaded.errors}),
            )

        req.context['dto'] = loaded
        req.context['_marshalled'] = True

    def process_response(self, req, resp, resource, req_succeeded):
        if isinstance(resp.body, MarshalResult):
            resp.content_type = 'application/json'
            resp.body = ujson.dumps(resp.body.data)



import os.path
from setuptools import setup, find_packages

cdir = os.path.abspath(os.path.dirname(__file__))

README = open(os.path.join(cdir, 'README.rst')).read()
CHANGELOG = open(os.path.join(cdir, 'CHANGELOG.rst')).read()

setup(
    name="falcon-helpers",
    description=('A few helpful tools to make working with the falcon '
                 'framework a real joy!'),
    long_description='\n\n'.join((README, CHANGELOG)),
    setup_requires=['setuptools_scm'],
    use_scm_version=True,
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'cryptography',
        'falcon',
        'falcon_multipart',
        'jinja2',
        'marshmallow',
        'marshmallow-sqlalchemy',
        'pyjwt',
        'sqlalchemy',
        'requests',
        'ujson',
        'wrapt',
    ],

    extras_require={
        's3': ['boto3'],
        'cli': ['click'],
        'dev': [
            'moto',
            'click',
            'pytest>=3.4',
            'raven',
        ]
    },
    zip_safe=True,
    license='BSD',
    author="Nick Zaccardi",
    author_email="nicholas.zaccardi@gmail.com",
    url='https://gitlab.com/skosh/falcon-helpers',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
)



import logging

from falcon import util


log = logging.getLogger(__name__)


class MultiMiddleware:
    """MultiMiddleware allows dynamic addition of middlewares onto the main middleware stack

    NOTE(nZac): If you use falcon_helpers.app.API you get this by default and `API.add_middleware`
    is mapped to this object.

    NOTE(nZac): This middleware only supports independent middleware. All process responses will be
    run regardless of the corresponding process_request being run.

    """
    __slots__ = (
        'req_mw',
        'resc_mw',
        'resp_mw',
    )

    def __init__(self, middleware=None):
        self.req_mw = list()
        self.resc_mw = list()
        self.resp_mw = list()

        middleware = middleware if middleware else list()

        if not isinstance(middleware, list):
            self.add_middleware(middleware)
        else:
            for mw in middleware:
                self.add_middleware(mw)

    def add_middleware(self, mw):
        process_request = util.get_bound_method(mw, 'process_request')
        process_resource = util.get_bound_method(mw, 'process_resource')
        process_response = util.get_bound_method(mw, 'process_response')

        if process_request:
            self.req_mw.append(process_request)

        if process_resource:
            self.resc_mw.append(process_resource)

        if process_response:
            self.resp_mw.append(process_response)

    def process_request(self, req, resp):
        for mw in self.req_mw:
            log.debug(f'Running {mw.__self__.__class__.__name__} middleware process_request.')
            try:
                mw(req, resp)
            except Exception as e:
                log.debug(f'{mw.__self__.__class__.__name__} MW process_request raised exception.')
                raise

    def process_resource(self, req, resp, resource, params):
        for mw in self.resc_mw:
            log.debug(f'Running {mw.__self__.__class__.__name__} middleware process_resource.')
            try:
                mw(req, resp, resource, params)
            except Exception as e:
                log.debug(f'{mw.__self__.__class__.__name__} MW process_resource raised exception.')
                raise

    def process_response(self, req, resp, resource, resp_succeded):
        for mw in self.resp_mw:
            log.debug(f'Running {mw.__self__.__class__.__name__} middleware process_response.')

            try:
                mw(req, resp, resource, resp_succeded)
            except Exception as e:
                log.debug(f'{mw.__self__.__class__.__name__} MW process_resource raised exception.')
                raise
